#include <apps/graph_embedding_p.h>

GraphEmbeddingProver::
GraphEmbeddingProver(int ph, int b_size, int num_r, int input_size)
  : Prover(ph, b_size, num_r, input_size, PROVER_NAME)
{
  init_state();
}

void GraphEmbeddingProver::
init_state()
{
  num_bits_in_prime = 320;
  num_bits_in_input = 32;
  crypto_in_use = CRYPTO_ELGAMAL;
  png_in_use = PNG_CHACHA;
 
  Prover::init_state();
  
  num_lin_pcp_queries = NUM_LIN_PCP_QUERIES;
  ConstraintGenerator cg(prime);
  cg.init_size();

  proof.graph_embedding_proof(cg, m);

  num_variables = cg.variables_size();
  num_variables2 = num_variables*num_variables;
  
  
  alloc_init_vec(&F1, num_variables);
  alloc_init_vec(&F2, num_variables2);
  alloc_init_vec(&qF1, num_variables);
  alloc_init_vec(&output, expansion_factor);
  alloc_init_vec(&f1_commitment, expansion_factor*num_variables);
  alloc_init_vec(&f2_commitment, expansion_factor*num_variables2);
  alloc_init_vec(&f1_q1, num_variables);
  alloc_init_vec(&f2_q1, num_variables2);

  alloc_init_vec(&f_answers, num_lin_pcp_queries);
  alloc_init_scalar(answer);

  F_ptrs.clear(); 
  F_ptrs.push_back(F1);
  F_ptrs.push_back(F2);
  f_q_ptrs.clear();
  f_q_ptrs.push_back(f1_q1);
  f_q_ptrs.push_back(f2_q1);
  
  find_cur_qlengths();
}

void GraphEmbeddingProver::
find_cur_qlengths()
{
  sizes.clear();
  sizes.push_back(num_variables);
  sizes.push_back(num_variables2);
  
  qquery_sizes.clear();
  qquery_sizes.push_back(num_variables);
  qquery_sizes.push_back(num_variables);
  qquery_sizes.push_back(num_variables2);
  qquery_sizes.push_back(num_variables);
  qquery_sizes.push_back(num_variables2);
  qquery_f_ptrs.clear();
  qquery_f_ptrs.push_back(f1_q1);
  qquery_f_ptrs.push_back(f1_q1);
  qquery_f_ptrs.push_back(f2_q1);
  qquery_f_ptrs.push_back(f1_q1);
  qquery_f_ptrs.push_back(f2_q1);
  qquery_F_ptrs.clear();
  qquery_F_ptrs.push_back(F1);
  qquery_F_ptrs.push_back(F1);
  qquery_F_ptrs.push_back(F2);
  qquery_F_ptrs.push_back(F1);
  qquery_F_ptrs.push_back(F2);
}

void GraphEmbeddingProver::
computation_graph_embedding()
{
  //TODO: implement me
  //Given that the first input_file_size elements of qF1 are filled with the values from 
  //the input file, this method should set elements input_file_size through input_file_size + output_file_size - 1
  //with the elements of the output for the computation.

  int maxPathLen = 2; //number of segments
  int edges = m-1; //whatever.
  int vertices = m;
  output_file_size = (vertices + edges * (maxPathLen+1))*2; //Correct this.
  int count = 0;
  int VERTICES = count;
  for(int i = 0; i < vertices; i++){
    mpq_set_si(qF1[count++], i, 1);
    mpq_set_si(qF1[count++], i, 1);
  }
  for(int e = 0; e < edges; e++){
    for(int i = 0; i < maxPathLen; i++){
      mpq_set_si(qF1[count], i, 4);
      mpq_add(qF1[count], qF1[count], qF1[VERTICES + (e)*2]);
      count++;
      mpq_set_si(qF1[count], i, 4);
      mpq_add(qF1[count], qF1[count], qF1[VERTICES + (e)*2 + 1]);
      count++;
    }
    mpq_set(qF1[count++], qF1[VERTICES + (e+1)*2]);
    mpq_set(qF1[count++], qF1[VERTICES + (e+1)*2 + 1]);
  }
}


//PROVER's CODE
void GraphEmbeddingProver::
prover_computation_commitment()
{
  // execute the computation
  load_vector(expansion_factor*num_variables, f1_commitment, (char *)"f1_commitment_query", FOLDER_WWW_DOWNLOAD);
  load_vector(expansion_factor*num_variables2, f2_commitment, (char *)"f2_commitment_query", FOLDER_WWW_DOWNLOAD);
  
  for (int i=batch_start; i<=batch_end; i++)
  {
    m_comp.begin_with_init(); 
    
    //The first input_file_size elements of qF1 are filled with the content of the input file.
    input_file_size = 0; //TODO: Fill this with the correct number of numbers in the input
    snprintf(scratch_str, BUFLEN-1, "input_b_%d", i);
    load_vector(input_file_size, qF1 + 0, scratch_str, FOLDER_WWW_DOWNLOAD);

    //The next output_file_size elements of qF1 are filled with the output size.
    computation_graph_embedding();

    //The remainder of qF1 is filled with intermediate proof variables. These are autogenerated.
    ConstraintGenerator cg(prime);
    cg.init_p(qF1);

    proof.graph_embedding_proof(cg, m);

    convert_to_z(num_variables, F1, qF1, prime);

    /* Useful for debugging. 
    for(int i = 0; i < num_variables; i++){
      gmp_printf("%Zd\n", F1[i]); 
    }
    */

    //Compute F2
    int index = 0;
    for (int a = 0; a < num_variables; a++){
      for(int b = 0; b < a; b++){
	mpz_set(F2[index], F2[b*num_variables + a]);
	index++;
      }
      for(int b = a; b < num_variables; b++){
	mpz_mul(F2[index], F1[a], F1[b]);
	index++;
      }
    }

    // start saving the state
    int output_file_size = m*2;
    snprintf(scratch_str, BUFLEN-1, "output_b_%d", i); 
    dump_vector(output_file_size, qF1 + input_file_size, scratch_str, FOLDER_WWW_DOWNLOAD);

    snprintf(scratch_str, BUFLEN-1, "outputz_b_%d", i); 
    dump_vector(output_file_size, F1 + input_file_size, scratch_str, FOLDER_WWW_DOWNLOAD);
    m_comp.end();
    
    snprintf(scratch_str, BUFLEN-1, "f1_assignment_vector_b_%d", i); 
    dump_vector(num_variables, F1, scratch_str, FOLDER_WWW_DOWNLOAD);
    
    snprintf(scratch_str, BUFLEN-1, "f2_assignment_vector_b_%d", i); 
    dump_vector(num_variables2, F2, scratch_str, FOLDER_WWW_DOWNLOAD);
  }

  for (int i=batch_start; i<=batch_end; i++)
  { 
    if (i == 0)
      m_proof_work.begin_with_init();
    else
      m_proof_work.begin_with_history();

    snprintf(scratch_str, BUFLEN-1, "f1_assignment_vector_b_%d", i); 
    load_vector(num_variables, F1, scratch_str, FOLDER_WWW_DOWNLOAD);
    
    snprintf(scratch_str, BUFLEN-1, "f2_assignment_vector_b_%d", i); 
    load_vector(num_variables2, F2, scratch_str, FOLDER_WWW_DOWNLOAD);

    if (crypto_in_use == CRYPTO_ELGAMAL)
      v->dot_product_enc(num_variables, f1_commitment, F1, output[0], output[1]);
    
    snprintf(scratch_str, BUFLEN-1, "f1_commitment_answer_b_%d", i); 
    dump_vector(expansion_factor, output, scratch_str, FOLDER_WWW_DOWNLOAD);
  
    if (crypto_in_use == CRYPTO_ELGAMAL)
      v->dot_product_enc(num_variables2, f2_commitment, F2, output[0], output[1]);
    
    snprintf(scratch_str, BUFLEN-1, "f2_commitment_answer_b_%d", i); 
    dump_vector(expansion_factor, output, scratch_str, FOLDER_WWW_DOWNLOAD);
    m_proof_work.end();
  }

}

void GraphEmbeddingProver::deduce_answers()
{
  //q1, q2, q4 => q3. 
  mpz_init_set_ui(f_answers[Q3], 0);
  mpz_t tmp;
  mpz_init_set_ui(tmp, 1);
  mpz_mul(tmp, f_answers[Q1], f_answers[Q2]);
  mpz_add(f_answers[Q3], tmp, f_answers[Q4]);
}

