#include <apps/polyeval_d2_p.h>

PolyEvalD2Prover::
PolyEvalD2Prover(int ph, int b_size, int num_r, int input_size)
  : Prover(ph, b_size, num_r, input_size, PROVER_NAME) {
  init_state();
}

void PolyEvalD2Prover::
init_state() {
  num_bits_in_prime = 128;
  num_bits_in_input = 32;
  crypto_in_use = CRYPTO_ELGAMAL;
  png_in_use = PNG_CHACHA;

  Prover::init_state();

  num_lin_pcp_queries = NUM_LIN_PCP_QUERIES;

  num_coefficients = (m * m + 3 * m)/2 + 1;
  alloc_init_vec(&F1, m);
  alloc_init_vec(&F2, m*m);
  alloc_init_vec(&output, expansion_factor);
  alloc_init_vec(&variables, m);
  alloc_init_vec(&coefficients, m*m+m+1);
  alloc_init_vec(&f1_commitment, expansion_factor*m);
  alloc_init_vec(&f1_consistency, expansion_factor*m);
  alloc_init_vec(&f1_q, m);
  alloc_init_vec(&f1_q2, m);
  alloc_init_vec(&f1_q3, m);
  alloc_init_vec(&f2_commitment, expansion_factor*m*m);
  alloc_init_vec(&f2_consistency, expansion_factor*m*m);
  alloc_init_vec(&f2_q, m*m);
  alloc_init_vec(&f2_q2, m*m);
  alloc_init_vec(&f2_q3, m*m);
  alloc_init_vec(&alpha, m+1);
  alloc_init_vec(&f_answers, num_lin_pcp_queries);
  alloc_init_scalar(answer);
  alloc_init_scalar(neg);

  F_ptrs.clear();
  F_ptrs.push_back(F1);
  F_ptrs.push_back(F2);
  f_q_ptrs.clear();
  f_q_ptrs.push_back(f1_q);
  f_q_ptrs.push_back(f2_q);
  f_q2_ptrs.clear();
  f_q2_ptrs.push_back(f1_q2);
  f_q2_ptrs.push_back(f2_q2);

  find_cur_qlengths();
}

void PolyEvalD2Prover::
find_cur_qlengths() {
  sizes.clear();
  sizes.push_back(m);
  sizes.push_back(m*m);

  qquery_sizes.clear();
  qquery_sizes.push_back(m);
  qquery_sizes.push_back(m);
  qquery_sizes.push_back(m*m);
  qquery_sizes.push_back(m*m);
  qquery_sizes.push_back(m);
  qquery_sizes.push_back(m*m);
  qquery_f_ptrs.clear();
  qquery_f_ptrs.push_back(f1_q);
  qquery_f_ptrs.push_back(f1_q);
  qquery_f_ptrs.push_back(f2_q);
  qquery_f_ptrs.push_back(f2_q);
  qquery_f_ptrs.push_back(f1_q);
  qquery_f_ptrs.push_back(f2_q);
  qquery_F_ptrs.clear();
  qquery_F_ptrs.push_back(F1);
  qquery_F_ptrs.push_back(F1);
  qquery_F_ptrs.push_back(F2);
  qquery_F_ptrs.push_back(F2);
  qquery_F_ptrs.push_back(F1);
  qquery_F_ptrs.push_back(F2);

  qquery_q_ptrs.clear();
  qquery_q_ptrs.push_back(Q1);
  qquery_q_ptrs.push_back(Q2);
  qquery_q_ptrs.push_back(Q3);
  qquery_q_ptrs.push_back(Q4);
  qquery_q_ptrs.push_back(Q5);
  qquery_q_ptrs.push_back(Q6);
}

// COMPUTATION
void PolyEvalD2Prover::
computation_assignment(mpz_t output) {
  // first, compute the quadratic assignment, F2 = \op{m}{m}
  int index, index2;
  for (int i=0; i<m; i++) {
    for (int j=i+1; j<m; j++) {
      index = m*i + j;
      index2 = m*j+i;
      mpz_mod(F2[index2], F2[index2], prime);
      mpz_set(F2[index], F2[index2]);
    }
  }
}

void PolyEvalD2Prover::
computation_polyeval(mpz_t output) {
  // first, compute the quadratic assignment, F2 = \op{m}{m}
  int index, index2;
  for (int i=0; i<m; i++) {
    mpz_set(F1[i], variables[i]);
    for (int j=0; j<=i; j++) {
      index = m*i+j;
      mpz_mul(F2[index], variables[i], variables[j]);
    }
  }

  // now compute the output; first the quadratic term
  mpz_t temp;
  mpz_init(temp);
  int k = 0;
  for (int i=0; i<m; i++) {
    for (int j=0; j<=i; j++) {
      int index = m*i+j;
      mpz_mul(temp, coefficients[k], F2[index]);
      mpz_add(output, output, temp);
      k++;
    }
  }

  // now the linear term
  for (int i=0; i<m; i++) {
    mpz_mul(temp, coefficients[k], F1[i]);
    mpz_add(output, output, temp);
    k++;
  }

  // now the constant term
  mpz_add(output, output, coefficients[k]);
  mpz_mod(output, output, prime);
}

//PROVER's CODE
void PolyEvalD2Prover::
prover_computation_commitment() {
  // init prover
  uint32_t s1 = m;
  uint32_t s2 = m*m;
  load_vector(expansion_factor*s1, f1_commitment, (char *)"f1_commitment_query", FOLDER_WWW_DOWNLOAD);
  load_vector(expansion_factor*s2, f2_commitment, (char *)"f2_commitment_query", FOLDER_WWW_DOWNLOAD);

  // execute the computation
  m_computation.begin_with_init();
  load_vector((m*m+3*m)/2+1, coefficients, (char *)"input0", FOLDER_WWW_DOWNLOAD);
  m_computation.end();

  for (int i=batch_start; i<=batch_end; i++) {
    m_computation.begin_with_history();
    //for (int k=0; k<INNER_LOOP_SMALL; k++)
    {
      snprintf(scratch_str, BUFLEN-1, "input1_b_%d", i);
      load_vector(m, variables, scratch_str, FOLDER_WWW_DOWNLOAD);

      for (int j=0; j<expansion_factor; j++)
        mpz_set_ui(output[j], 0);

      computation_polyeval(output[0]);

      // start saving the state
      snprintf(scratch_str, BUFLEN-1, "output_b_%d", i);
      dump_scalar(output[0], scratch_str, FOLDER_WWW_DOWNLOAD);
    }
    m_computation.end();

    computation_assignment(output[0]);
    snprintf(scratch_str, BUFLEN-1, "f1_assignment_vector_b_%d", i);
    dump_vector(m, F1, scratch_str, FOLDER_WWW_DOWNLOAD);

    snprintf(scratch_str, BUFLEN-1, "f2_assignment_vector_b_%d", i);
    dump_vector(m*m, F2, scratch_str, FOLDER_WWW_DOWNLOAD);
  }

  for (int i=batch_start; i<=batch_end; i++) {
    if (i == 0)
      m_answer_queries.begin_with_init();
    else
      m_answer_queries.begin_with_history();

    snprintf(scratch_str, BUFLEN-1, "f1_assignment_vector_b_%d", i);
    load_vector(m, F1, scratch_str, FOLDER_WWW_DOWNLOAD);

    snprintf(scratch_str, BUFLEN-1, "f2_assignment_vector_b_%d", i);
    load_vector(m*m, F2, scratch_str, FOLDER_WWW_DOWNLOAD);

    if (crypto_in_use == CRYPTO_ELGAMAL)
      v->dot_product_enc(m, f1_commitment, F1, dotp[0], dotp[1]);

    if (crypto_in_use == CRYPTO_ELGAMAL)
      v->dot_product_enc(m*m, f2_commitment, F2, output[0], output[1]);

    v->add_enc(dotp[0], dotp[1], dotp[0], dotp[1], output[0], output[1]);

    snprintf(scratch_str, BUFLEN-1, "f_commitment_answer_b_%d", i);
    dump_vector(expansion_factor, dotp, scratch_str, FOLDER_WWW_DOWNLOAD);

    m_answer_queries.end();
  }
}

void PolyEvalD2Prover::deduce_queries() {
  for (int rho=0; rho<num_repetitions; rho++) {
    snprintf(scratch_str, BUFLEN-1, "q1_qquery_r_%d", rho);
    load_vector(m, f1_q, scratch_str, FOLDER_WWW_DOWNLOAD);

    snprintf(scratch_str, BUFLEN-1, "q2_qquery_r_%d", rho);
    load_vector(m, f1_q2, scratch_str, FOLDER_WWW_DOWNLOAD);

    snprintf(scratch_str, BUFLEN-1, "q4_qquery_r_%d", rho);
    load_vector(m*m, f2_q, scratch_str, FOLDER_WWW_DOWNLOAD);

    v->outer_product(m, f1_q, m, f1_q2, f2_q, prime, f2_q2);
    snprintf(scratch_str, BUFLEN-1, "q3_qquery_r_%d", rho);
    dump_vector(m*m, f2_q2, scratch_str);
  }

  /*  mpz_init_set_ui(f_answers[Q3], 0);
    mpz_t tmp;
    mpz_init_set_ui(tmp, 1);
    mpz_mul(tmp, f_answers[Q1], f_answers[Q2]);
    mpz_add(f_answers[Q3], tmp, f_answers[Q4]);
  */
}
